#pragma once
#include "includes.h"
#include "scene_elements.h"
#include "intersect.h"

void vec_min(double vec[3], double bound_min[3]) {
	for (char i=0;i<3;++i) bound_min[i] = min(vec[i],bound_min[i]);
}
void vec_max(double vec[3], double bound_max[3]) {
	for (char i=0;i<3;++i) bound_max[i] = max(vec[i],bound_max[i]);
}
void vec_min_max(double vec[3], double bound_min[3], double bound_max[3]) {
	vec_min(vec,bound_min);
	vec_max(vec,bound_max);
}
void reset_bounds(double bound_min[3], double bound_max[3]) {
	bound_min[0] = bound_min[1] = bound_min[2] = MAX_DOUBLE;
	bound_max[0] = bound_max[1] = bound_max[2] = MIN_DOUBLE;
}
void determine_bounds(double bound_min[3], double bound_max[3], Object* obj, bool reset=true) {
	if (reset) reset_bounds(bound_min,bound_max);
	if (obj==NULL) return;
	switch (obj->type) {
		case (SCENE_OBJECT_TYPE_POINT):
			vec_min_max(obj->position,bound_min,bound_max);
			break;
		case (SCENE_OBJECT_TYPE_SPHERE):
			bound_min[0] = min( bound_min[0], obj->position[0]-((Sphere*)(obj))->radius );
			bound_min[1] = min( bound_min[1], obj->position[1]-((Sphere*)(obj))->radius );
			bound_min[2] = min( bound_min[2], obj->position[2]-((Sphere*)(obj))->radius );
			bound_max[0] = max( bound_max[0], obj->position[0]+((Sphere*)(obj))->radius );
			bound_max[1] = max( bound_max[1], obj->position[1]+((Sphere*)(obj))->radius );
			bound_max[2] = max( bound_max[2], obj->position[2]+((Sphere*)(obj))->radius );
			break;
		case (SCENE_OBJECT_TYPE_PLANE):
			//Do nothing because we're going to test all planes anyway
			break;
		case (SCENE_OBJECT_TYPE_TRIANGLE):
			vec_min_max(             obj  ->position, bound_min,bound_max);
			vec_min_max(((Triangle*)(obj))->position2,bound_min,bound_max);
			vec_min_max(((Triangle*)(obj))->position3,bound_min,bound_max);
			break;
		default:
			throw glLibError("Unrecognized type!");
	}
}
void determine_bounds(double bound_min[3], double bound_max[3], list<Object*>* objs, bool reset=true) {
	if (reset) reset_bounds(bound_min,bound_max);
	for (list<Object*>::const_iterator iter=objs->begin();iter!=objs->end();++iter) {
		Object* obj = *iter;
		determine_bounds(bound_min,bound_max,obj,false);
	}
}

class BoundingVolume {
	friend bool compare_coords_x(BoundingVolume* a, BoundingVolume* b);
	friend bool compare_coords_y(BoundingVolume* a, BoundingVolume* b);
	friend bool compare_coords_z(BoundingVolume* a, BoundingVolume* b);
	friend class NodeBVH;
	private:
		double _min[3];
		double _max[3];
		double _center[3];
		Object* object;
		bool _is_infinite;
	public:
		BoundingVolume() {
			reset_bounds(_min,_max); _is_infinite = true;
			object = NULL;
			recalculate_average();
		}
		BoundingVolume(Object* obj) {
			determine_bounds(_min,_max,obj); //initializes to max/min int
			if (obj==NULL) _is_infinite = true;
			else _is_infinite = false;
			object = obj;
			recalculate_average();
		}
		BoundingVolume(list<Object*>* bvs) {
			reset_bounds(_min,_max);
			object = NULL;
			_is_infinite = true;
			for (list<Object*>::const_iterator iter=bvs->begin();iter!=bvs->end();++iter) {
				BoundingVolume temp(*iter);
				add_min(temp._min);
				add_max(temp._max);
				_is_infinite = false;
			}
			//recalculate_average(); //add_min and add_max do this
		}
		BoundingVolume(list<BoundingVolume*>* bvs) {
			reset_bounds(_min,_max);
			object = NULL;
			_is_infinite = true;
			for (list<BoundingVolume*>::const_iterator iter=bvs->begin();iter!=bvs->end();++iter) {
				add_min((*iter)->_min);
				add_max((*iter)->_max);
				_is_infinite = false;
			}
			//recalculate_average(); //add_min and add_max do this
		}
		BoundingVolume(BoundingVolume* a, BoundingVolume* b) {
			reset_bounds(_min,_max);
			object = NULL;
			add_min(a->_min); add_max(a->_max);
			add_min(b->_min); add_max(b->_max);
			_is_infinite = a->_is_infinite||b->_is_infinite;
			//recalculate_average(); //add_min and add_max do this
		}

		inline double get_volume(void) {
			return _is_infinite ? MAX_DOUBLE : (_max[0]-_min[0])*(_max[1]-_min[1])*(_max[2]-_min[2]);
		}
		void add_min(double _min[3]) {
			for (char i=0;i<3;++i) this->_min[i] = min(this->_min[i],_min[i]);
			_is_infinite = false;
			recalculate_average();
		}
		void add_max(double _max[3]) {
			for (char i=0;i<3;++i) this->_max[i] = max(this->_max[i],_max[i]);
			_is_infinite = false;
			recalculate_average();
		}
		static double get_overlap(BoundingVolume* a, BoundingVolume* b) {
			BoundingVolume overlap = BoundingVolume();
			overlap.add_min(a->_min);
			overlap.add_min(b->_min);
			overlap.add_max(a->_max);
			overlap.add_max(b->_max);
			return overlap.get_volume();
		}
		void print_float(double val) {
			if (val>=10.0) printf("+INF");
			else if (val<=-10.0) printf("-INF");
			else printf("%+1.1f",val);
		}
		void print_bounds(void) {
			printf("(");print_float(_min[0]);printf(",");print_float(_max[0]);printf(") ");
			printf("(");print_float(_min[1]);printf(",");print_float(_max[1]);printf(") ");
			printf("(");print_float(_min[2]);printf(",");print_float(_max[2]);printf(")");
		}
	private:
		double intersect(double origin[3], double recip_direction[3]) {
			//http://www.cs.utah.edu/~awilliam/box/box.pdf
			double tmin, tmax, tymin, tymax, tzmin, tzmax;

			if (recip_direction[0]<0.0) {
				tmin  = (_max[0] - origin[0]) * recip_direction[0];
				tmax  = (_min[0] - origin[0]) * recip_direction[0];
			}
			else {
				tmin  = (_min[0] - origin[0]) * recip_direction[0];
				tmax  = (_max[0] - origin[0]) * recip_direction[0];
			}
			if (recip_direction[1]<0.0) {
				tymin = (_max[1] - origin[1]) * recip_direction[1];
				tymax = (_min[1] - origin[1]) * recip_direction[1];
			}
			else {
				tymin = (_min[1] - origin[1]) * recip_direction[1];
				tymax = (_max[1] - origin[1]) * recip_direction[1];
			}

			if (tmin>tymax||tymin>tmax) return 0.0;
			if (tymin>tmin) tmin = tymin;
			if (tymax<tmax) tmax = tymax;

			if (recip_direction[2]<0.0) {
				tzmin = (_max[2] - origin[2]) * recip_direction[2];
				tzmax = (_min[2] - origin[2]) * recip_direction[2];
			}
			else {
				tzmin = (_min[2] - origin[2]) * recip_direction[2];
				tzmax = (_max[2] - origin[2]) * recip_direction[2];
			}

			if ( (tmin > tzmax) || (tzmin > tmax) ) return 0.0;
			if (tzmin > tmin) tmin = tzmin;
			if (tzmax < tmax) tmax = tzmax;

			return (tmin<0.0) ? tmax : tmin; //tmin<t1 && tmax>t0;
		}
		void recalculate_average(void) {
			//Average position
			glLibMathVec::add3(_min,_max,_center);
			glLibMathVec::scale3(0.5,_center);
		}
};

bool compare_coords_x(BoundingVolume* a, BoundingVolume* b) {
	return (a->_center[0]<b->_center[0]);
}
bool compare_coords_y(BoundingVolume* a, BoundingVolume* b) {
	return (a->_center[1]<b->_center[1]);
}
bool compare_coords_z(BoundingVolume* a, BoundingVolume* b) {
	return (a->_center[2]<b->_center[2]);
}

double intersect_object(list<Object*>::const_iterator iter, double direction[3], double origin[3]) {
	switch ((*iter)->type) {
		case    SCENE_OBJECT_TYPE_POINT: {
			return 0.0;
			break;
		}
		case   SCENE_OBJECT_TYPE_SPHERE: {
			Sphere* object = (Sphere*)(*iter);
			return intersect_sphere(direction,origin,object->position,object->radius);
			break;
		}
		case    SCENE_OBJECT_TYPE_PLANE: {
			Plane* object = (Plane*)(*iter);
			return intersect_plane(direction,origin,object->position,object->normal);
			break;
		}
		case SCENE_OBJECT_TYPE_TRIANGLE: {
			Triangle* object = (Triangle*)(*iter);
			return intersect_triangle(direction,origin,object->position,object->position2,object->position3);
			break;
		}
		default: { throw glLibError("Not implemented!"); break; }
	}
	return 0.0; //not possible
}

class NodeBVH {
	public:
		int depth;
		NodeBVH *child1, *child2;
		list<Object*>* elements;
		BoundingVolume* bv;
		bool is_leaf;
	public:
		NodeBVH(list<BoundingVolume*>* bounding_volumes, int depth, int max_depth) {
			this->depth = depth;
			//print_start(); printf("Adding node at depth %d with %d to store!\n",depth,bounding_volumes->size());
			//print_start(); printf("| \n");
			if (bounding_volumes->size()==1 || depth==max_depth) {
				is_leaf = true;
				//print_start(); printf("| This is a leaf node.\n");
				child1 = child2 = NULL;
				elements = new list<Object*>();
				for (list<BoundingVolume*>::const_iterator iter=bounding_volumes->begin();iter!=bounding_volumes->end();++iter) {
					//print_start(); printf("| Adding bounding volume "); (*iter)->print_bounds(); printf(".\n");
					elements->push_back((*iter)->object);
				}
				//print_start(); printf("| \n");
				bv = new BoundingVolume(elements);
			}
			else {
				is_leaf = false;
				//print_start(); printf("| This is an ordinary node.\n");
				//print_start(); printf("| Attempting to find the two smallest BVs.\n");

				elements = NULL;
				list<BoundingVolume*> list_x_1, list_x_2, list_y_1, list_y_2, list_z_1, list_z_2;

				bounding_volumes->sort(compare_coords_x);
				double best_overlap_x = optimize(bounding_volumes, &list_x_1,&list_x_2);
				/*print_start(); printf("| In x, the best overlap is %f, corresponding to:\n",best_overlap_x);
				print_start(); printf("| | List 1  ( left): "); BoundingVolume(&list_x_1).print_bounds(); printf("\n");
				for (list<BoundingVolume*>::const_iterator iter=list_x_1.begin();iter!=list_x_1.end();++iter) {
					print_start(); printf("| | | "); (*iter)->print_bounds(); printf(".\n");
				}
				print_start(); printf("| | List 2  (right): "); BoundingVolume(&list_x_2).print_bounds(); printf("\n");
				for (list<BoundingVolume*>::const_iterator iter=list_x_2.begin();iter!=list_x_2.end();++iter) {
					print_start(); printf("| | | "); (*iter)->print_bounds(); printf(".\n");
				}*/

				bounding_volumes->sort(compare_coords_y);
				double best_overlap_y = optimize(bounding_volumes, &list_y_1,&list_y_2);
				/*print_start(); printf("| In y, the best overlap is %f, corresponding to:\n",best_overlap_x);
				print_start(); printf("| | List 1 (bottom): "); BoundingVolume(&list_y_1).print_bounds(); printf("\n");
				for (list<BoundingVolume*>::const_iterator iter=list_y_1.begin();iter!=list_y_1.end();++iter) {
					print_start(); printf("| | | "); (*iter)->print_bounds(); printf(".\n");
				}
				print_start(); printf("| | List 2 (   top): "); BoundingVolume(&list_y_2).print_bounds(); printf("\n");
				for (list<BoundingVolume*>::const_iterator iter=list_y_2.begin();iter!=list_y_2.end();++iter) {
					print_start(); printf("| | | "); (*iter)->print_bounds(); printf(".\n");
				}*/

				bounding_volumes->sort(compare_coords_z);
				double best_overlap_z = optimize(bounding_volumes, &list_z_1,&list_z_2);
				/*print_start(); printf("| In z, the best overlap is %f, corresponding to:\n",best_overlap_x);
				print_start(); printf("| | List 1 (front?): "); BoundingVolume(&list_z_1).print_bounds(); printf("\n");
				for (list<BoundingVolume*>::const_iterator iter=list_z_1.begin();iter!=list_z_1.end();++iter) {
					print_start(); printf("| | | "); (*iter)->print_bounds(); printf(".\n");
				}
				print_start(); printf("| | List 2 ( back?): "); BoundingVolume(&list_z_2).print_bounds(); printf("\n");
				for (list<BoundingVolume*>::const_iterator iter=list_z_2.begin();iter!=list_z_2.end();++iter) {
					print_start(); printf("| | | "); (*iter)->print_bounds(); printf(".\n");
				}*/

				if (best_overlap_x<best_overlap_y&&best_overlap_x<best_overlap_z) { //best = x
					//print_start(); printf("| x is the best!\n");
					child1 = new NodeBVH(&list_x_1,depth+1,max_depth);
					child2 = new NodeBVH(&list_x_2,depth+1,max_depth);
				}
				else if (best_overlap_y<best_overlap_x&&best_overlap_y<best_overlap_z) { //best = y
					//print_start(); printf("| y is the best!\n");
					child1 = new NodeBVH(&list_y_1,depth+1,max_depth);
					child2 = new NodeBVH(&list_y_2,depth+1,max_depth);
				}
				else { //best = z
					//print_start(); printf("| z is the best!\n");
					child1 = new NodeBVH(&list_z_1,depth+1,max_depth);
					child2 = new NodeBVH(&list_z_2,depth+1,max_depth);
				}

				bv = new BoundingVolume(child1->bv,child2->bv);
			}
			//print_start(); printf("This BV is at "); bv->print_bounds(); printf(".\n");
			//print_start(); printf("\n");
		}
		~NodeBVH() {
			if (child1!=NULL) delete child1;
			if (child2!=NULL) delete child2;
			if (elements!=NULL) delete elements;
			delete bv;
		}

		unsigned int intersect(double origin[3], double direction[3], double recip_direction[3], double* distance, Object** object_hit) {
			if (bv->intersect(origin,recip_direction)==0.0) return 0;

			unsigned int total;
			if (elements==NULL) {
				total = 0;
				total += child1->intersect(origin,direction,recip_direction,distance,object_hit);
				total += child2->intersect(origin,direction,recip_direction,distance,object_hit);
			}
			else {
				for (list<Object*>::const_iterator iter=elements->begin(); iter!=elements->end(); ++iter) {
					double intersect = intersect_object(iter,direction,origin);
					if (intersect<=0.0) continue;
					if (intersect<*distance) {
						*distance = intersect;
						*object_hit = *iter;
					}
				}
				total = elements->size();
			}
			return total;
		}

		double get_avg_in_leaf(void) {
			return (is_leaf) ? (double)(elements->size()) : (child1->get_avg_in_leaf()+child2->get_avg_in_leaf())/2.0;
		}
		unsigned int get_max_in_leaf(void) {
			return (is_leaf) ? elements->size() : max(child1->get_max_in_leaf(),child2->get_max_in_leaf());
		}
		double get_avg_depth(void) {
			return (is_leaf) ? (double)(depth) : (child1->get_avg_depth()+child2->get_avg_depth())/2.0;
		}
		unsigned int get_max_depth(void) {
			return (is_leaf) ? depth : max(child1->get_max_depth(),child2->get_max_depth());
		}

		void print(void) {
			print_start(); printf("Begin node:\n");
			if (elements==NULL) {
				print_start(); printf("Child 1:\n");
				child1->print();
				print_start(); printf("End child 1:\n");
				print_start(); printf("Child 2:\n");
				child2->print();
				print_start(); printf("End child 2:\n");
			}
			else {
				for (list<Object*>::const_iterator iter=elements->begin();iter!=elements->end();++iter) {
					Object* obj = *iter;
					print_start(); printf("  Object \"%s\" at ",obj->name->c_str());
					glLibMathVec::print3(obj->position);
				}
			}
			print_start(); printf("End node\n");
		}
	private:
		inline void print_start(void) {
			for (int i=0;i<depth;++i) printf("|   ");
		}
		double optimize(list<BoundingVolume*>* data, list<BoundingVolume*>* l1, list<BoundingVolume*>* l2) {
			list<BoundingVolume*>::const_iterator iter = data->begin();
			unsigned int i = 0;
			for (;i<data->size()/2;++i) {
				l1->push_back(*iter);
				++iter;
			}
			for (;i<data->size();++i) {
				l2->push_back(*iter);
				++iter;
			}
			return BoundingVolume(l1).get_volume() + BoundingVolume(l2).get_volume();
			
			/*//Optimize to find the two AABB that have the smallest summed volume

			for (list<BoundingVolume*>::const_iterator iter=data->begin();iter!=data->end();++iter) {
				BoundingVolume* bv = *iter;
				l2->push_back(bv);
			}

			BoundingVolume bv1,bv2(l2);
			print_start(); printf("Beginning with "); bv2.print_bounds(); printf("\n");

			double total_previous = bv2.get_volume();

			while (l2->size()>0) {
				print_start(); printf("Starting iteration\n");
				print_start(); printf("  Previous sum: %f\n",total_previous);

				BoundingVolume* pivot = l2->front();
				l1->push_back(pivot);
				l2->pop_front();
				bv1.add_min(pivot->_min);
				bv1.add_max(pivot->_max);
				bv2 = BoundingVolume(l2);

				print_start(); printf("  List 1: "); BoundingVolume(l1).print_bounds(); printf("\n");
				for (list<BoundingVolume*>::const_iterator iter=l1->begin();iter!=l1->end();++iter) {
					print_start(); printf("    "); (*iter)->print_bounds(); printf(".\n");
				}
				print_start(); printf("  List 2: "); BoundingVolume(l2).print_bounds(); printf("\n");
				for (list<BoundingVolume*>::const_iterator iter=l2->begin();iter!=l2->end();++iter) {
					print_start(); printf("    "); (*iter)->print_bounds(); printf(".\n");
				}

				double total = bv1.get_volume() + bv2.get_volume();

				print_start(); printf("  Sum: %f = %f + %f\n",total,bv1.get_volume(),bv2.get_volume());

				if (total>total_previous) {
					//passed optimization point or used all elements; go back and then return
					BoundingVolume* pivot = l1->back();
					l2->push_front(pivot);
					l1->pop_back();
					print_start(); printf("Passed optimization point; reverting and returning.\n");
					return total_previous;
				}
				total_previous = total;
				print_start(); printf("  Ending iteration\n");
			}
			print_start(); printf("Surpassed all elements.\n");
			return bv1.get_volume(); //took all the elements out of l2 and put them into l1*/
		}
};
class WeightedBVH {
	private:
		double _min[3];
		double _max[3];
		list<Object*>* _objects;
		list<Plane*>* _planes;
		list<BoundingVolume*>* _bounding_volumes;
		NodeBVH* tree;
		unsigned int _max_allowed_depth;
		unsigned int _num_objects_triangles;
		unsigned int _num_objects_spheres;
		unsigned int _num_objects_planes;
	public:
		WeightedBVH(list<Object*>* objects, unsigned int max_allowed_depth) {
			_max_allowed_depth = max_allowed_depth;
			_objects = objects;
			_planes = new list<Plane*>();
			determine_bounds(_min,_max,objects);

			_num_objects_triangles = 0;
			_num_objects_spheres = 0;
			_num_objects_planes = 0;
			for (list<Object*>::const_iterator iter=objects->begin(); iter!=objects->end(); ++iter) {
				switch ((*iter)->type) {
					case SCENE_OBJECT_TYPE_TRIANGLE:
						++_num_objects_triangles;
						break;
					case SCENE_OBJECT_TYPE_SPHERE:
						++_num_objects_spheres;
						break;
					case SCENE_OBJECT_TYPE_PLANE:
						++_num_objects_planes;
						_planes->push_back((Plane*)(*iter)); //can hit, but can't be optimized by BVH
						break;
					case SCENE_OBJECT_TYPE_POINT: //ignore; can't hit it
						break;
					default:
						throw glLibError("Unrecognized object type!");
				}
			}

			//printf("Got bounding volumes\n");
			_bounding_volumes = new list<BoundingVolume*>();
			for (list<Object*>::const_iterator iter=objects->begin();iter!=objects->end();++iter) {
				_bounding_volumes->push_back(new BoundingVolume(*iter));
				//printf("  ");_bounding_volumes->back()->print_bounds(); printf("\n");
			}

			tree = new NodeBVH(_bounding_volumes,0,_max_allowed_depth);
		}
		~WeightedBVH() {
			delete _planes;

			for (list<BoundingVolume*>::iterator iter=_bounding_volumes->begin();iter!=_bounding_volumes->end();++iter) {
				delete (*iter);
			}
			delete _bounding_volumes;

			delete tree;
		}

		unsigned int intersect(double origin[3], double direction[3], double* distance, Object** object_hit) {
			*object_hit = NULL;
			*distance = MAX_DOUBLE;
			double recip_direction[3] = {1.0/direction[0],1.0/direction[1],1.0/direction[2]};

			unsigned int tests = tree->intersect(origin,direction,recip_direction, distance,object_hit);
			for (list<Plane*>::const_iterator iter=_planes->begin(); iter!=_planes->end(); ++iter) {
				Plane* plane = *iter;
				double dist = intersect_plane(direction,origin,plane->position,plane->normal);
				if (dist>0.0&&dist<*distance) {
					*distance = dist;
					*object_hit = plane;
				}
				++tests;
			}
			return tests;
		}

		inline double get_avg_in_leaf(void) {
			return tree->get_avg_in_leaf();
		}
		inline unsigned int get_max_in_leaf(void) {
			return tree->get_max_in_leaf();
		}
		inline double get_avg_depth(void) {
			return tree->get_avg_depth();
		}
		inline unsigned int get_max_depth(void) {
			return tree->get_max_depth();
		}
		inline unsigned int get_num_objects(void) {
			return _objects->size();
		}
		inline unsigned int get_max_possible_depth(void) {
			return _max_allowed_depth;
		}
		inline unsigned int get_num_objects_triangles(void) {
			return _num_objects_triangles;
		}
		inline unsigned int get_num_objects_spheres(void) {
			return _num_objects_spheres;
		}
		inline unsigned int get_num_objects_planes(void) {
			return _num_objects_planes;
		}

		inline void print(void) {
			tree->print();
		}
	private:
};