﻿#pragma once
#include "../../stdafx.h"
#include "../glLibLocals.h"

#include "../glLibError.h"
#include "glLibMathFraction.h"

template <typename type_element, typename type_int> class glLibMathMatrix;

#ifdef GLLIB_ERROR_MODE_REPORT_ERRORS
template <typename type_element, typename type_int> void glLibInternal_check_3x3(glLibMathMatrix<type_element,type_int>* matrix) {
	if ((matrix->_rows==3)&&(matrix->_columns==3)) return;
	throw glLibError("Cannot compute operation; matrix must be 3x3.");
}
template <typename type_element, typename type_int> void glLibInternal_check_4x4(glLibMathMatrix<type_element,type_int>* matrix) {
	if ((matrix->_rows==4)&&(matrix->_columns==4)) return;
	throw glLibError("Cannot compute operation; matrix must be 4x4.");
}
template <typename type_element, typename type_int> void glLibInternal_check_3x3or4x4(glLibMathMatrix<type_element,type_int>* matrix) {
	if ((matrix->_rows==3)&&(matrix->_columns==3)) return;
	if ((matrix->_rows==4)&&(matrix->_columns==4)) return;
	throw glLibError("Cannot compute operation; matrix must be 3x3 or 4x4.");
}
#endif

//TODO: in-place transpose method, jacobi, inversion, copy_to(GLfloat[16]), operators :-(
//TODO: make work nicely with glLibFraction
//TODO: glLibMath -> glLibExtMath
template <typename type_element, typename type_int> class glLibMathMatrix {
	//Represents matrices internally as an array of rows:
	//.values[row][column]

	friend class glLibInternal_glLibView;
	template <typename type_element2, typename type_int2> friend void glLibInternal_check_3x3(glLibMathMatrix<type_element2,type_int2>* matrix);
	template <typename type_element2, typename type_int2> friend void glLibInternal_check_4x4(glLibMathMatrix<type_element2,type_int2>* matrix);
	template <typename type_element2, typename type_int2> friend void glLibInternal_check_3x3or4x4(glLibMathMatrix<type_element2,type_int2>* matrix);

	typedef glLibMathMatrix<type_element,type_int> type_matrix;

	//Heavily adapted from http://www.acims.arizona.edu/PUBLICATIONS/PDF/RajnikanthTemplates04.pdf
	public:
		type_element** values;
	private:
		type_int _rows; //TODO: numof_rows, numof_columns
		type_int _columns;
	private:
		void glLibInternal_init(type_int rows, type_int columns) {
			_rows = rows;
			_columns = columns;

			values = new type_element*[_rows];
			for (type_int i=0;i<_rows;++i) {
				values[i] = new type_element[_columns];
			}
		}
	public:
		//Make an empty, identity, zeroes, or ones, matrix
		glLibMathMatrix(type_int rows, type_int columns, glLibFlagMatrix flag) {
			glLibInternal_init(rows,columns);
			if      (flag==GLLIB_MATRIX_EMPTY);
			else if (flag==GLLIB_MATRIX_IDENTITY) {
				for (type_int i=0;i<_rows;++i) {
					for (type_int j=0;j<_columns;++j) {
						if (i==j) values[i][j] = 1;
						else values[i][j] = 0;
					}
				}
			}
			else if (flag==GLLIB_MATRIX_ZEROS) { //implicitly includes GLLIB_MATRIX_ZEROES, with "E"
				for (type_int i=0;i<_rows;++i) {
					for (type_int j=0;j<_columns;++j) {
						values[i][j] = 0;
					}
				}
			}
			#ifdef GLLIB_ERROR_MODE_REPORT_ERRORS
			else if (flag==GLLIB_MATRIX_ONES) {
			#else
			else {
			#endif
				for (type_int i=0;i<_rows;++i) {
					for (type_int j=0;j<_columns;++j) {
						values[j][j] = 1;
					}
				}
			}
			#ifdef GLLIB_ERROR_MODE_REPORT_ERRORS
			else {
				throw glLibError("Matrix type not recognized!");
			}
			#endif
		}
		//Make matrix from data
		glLibMathMatrix(type_int rows, type_int columns, type_element* data, glLibFlagMatrix row_column_flag) {
			glLibInternal_init(rows,columns);
			
			if (row_column_flag==GLLIB_MATRIX_ROW) {
				type_int index = 0;
				for (type_int i=0;i<_columns;++i) {
					for (type_int j=0;j<_rows;++j) {
						values[i][j] = data[index++];
					}
				}
			}
			#ifdef GLLIB_ERROR_MODE_REPORT_ERRORS
			else if (row_column_flag==GLLIB_MATRIX_COLUMN) {
			#else
			else {
			#endif
				//Matrix creation from array goes down columns instead of across rows.  
				//Good for direct and intuitive copying to and from OpenGL.
				type_int index = 0;
				for (type_int i=0;i<_columns;++i) {
					for (type_int j=0;j<_rows;++j) {
						values[j][i] = data[index++];
					}
				}
			}
			#ifdef GLLIB_ERROR_MODE_REPORT_ERRORS
			else {
				throw glLibError("Array interpretation not recognized!  Must be GLLIB_MATRIX_ROW or GLLIB_MATRIX_COLUMN.");
			}
			#endif
		}
		//Make matrix from data (array is interpreted as array of rows, just as internal state)
		glLibMathMatrix(type_int rows, type_int columns, type_element** data) {
			glLibInternal_init(rows,columns);

			for (type_int i=0;i<_rows;++i) {
				for (type_int j=0;j<_columns;++j) {
					values[i][j] = data[i][j];
				}
			}
		}
		//Make an augmented matrix
		//TODO: add overloads these two functions?
		glLibMathMatrix(type_matrix* mat1, type_matrix* mat2) {
			#ifdef GLLIB_ERROR_MODE_REPORT_ERRORS
			if (mat1->_rows!=mat2->_rows) throw glLibError("Row dimensions do not match!  Cannot construct augmented matrix!");
			#endif

			glLibInternal_init(mat1->_rows,mat1->_columns+mat2->_columns);

			for (type_int i=0;i<_rows;++i) {
				for (type_int j=0;j<mat1->_columns;++j) {
					values[i][j] = mat1->values[i][j];
				}
				for (type_int j=mat1->_columns;j<_columns;++j) {
					values[i][j] = mat2->values[i][j-mat1->_columns];
				}
			}
		}
		//Make a submatrix
		glLibMathMatrix(type_matrix* mat, type_int left, type_int right, type_int top, type_int bottom) {
			throw glLibError("Not implemented!");
		}
		//Copy from a matrix
		glLibMathMatrix(const type_matrix& a) {
			glLibInternal_init(a._rows,a._columns);

			for (type_int i=0;i<_rows;++i) {
				for (type_int j=0;j<_columns;++j) {
					values[i][j] = a.values[i][j];
				}
			}
		}
		~glLibMathMatrix() {
			for (type_int i=0;i<_rows;++i) {
				delete [] values[i];
			}
			delete [] values;
		}

		void print(void) {
			//TODO: make #ifdef to use unicode and non-unicode characters
			//TODO: format arguments
			//TODO: newline

			/*printf("\xC9"); //╔ 201, 2554
			for (type_int i=0; i<_columns; ++i) printf("         ");
			printf("\xBB\n"); //╗, 187, 2557*/
			printf("\xDA"); //┌
			for (type_int i=0; i<_columns; ++i) printf("         ");
			printf("\xBF\n"); //┐

			for (type_int i=0; i<_rows; ++i) {
				//printf("\xBA"); //║ 186, 2551
				printf("\xB3"); //│
				for (type_int j=0; j<_columns; ++j) {
					glLibInternal_print_element(values[i][j]);
				}
				//printf("\xBA\n"); //║ 186, 2551
				printf("\xB3\n"); //│
			}
			
			/*printf("\xC8"); //╚ 200, 255A
			for (type_int i=0; i<_columns; ++i) printf("         ");
			printf("\xBC\n"); //╝ 188, 255D*/
			printf("\xC0"); //└
			for (type_int i=0; i<_columns; ++i) printf("         ");
			printf("\xD9\n"); //┘
		}
	private:
		template <typename type_int2> void glLibInternal_print_element(glLibMathFraction<type_int2> value) {
			printf(" (");
			value.glLibInternal_print_numerator();
			printf("/");
			value.glLibInternal_print_denominator();
			printf(") ");
		}
		template <typename type_element2> void glLibInternal_print_element(type_element2 value) {
			printf("% 8.4f ",value);
			//printf("% f\t",values[i][j]);
		}
	public:

		//TODO: copy_from_OpenGL({GL_MODELVIEW|GL_PROJECTION|GL_TEXTURE|etc.})
		//TODO: copy_to_OpenGL({GL_MODELVIEW|GL_PROJECTION|GL_TEXTURE|etc.})

		//TODO: test these and set_at
		//TODO: row column, not column row (ray tracer uses this incorrectly, change)
		inline type_element get_at(type_int column, type_int row) { //column, row, corresponds to x, y
			return values[row][column];
		}
		inline type_element get_at(type_int n) { //Chooses the nth element (by mathematical standard; reads along rows)
			return values[n/_rows][n%_rows];
		}
		//TODO: set_at

		static void add(type_matrix* m1, type_matrix* m2, type_matrix* result) {
			#ifdef GLLIB_ERROR_MODE_REPORT_ERRORS
			if (m1->_rows!=m2->_rows||m1->_columns!=m2->_columns) throw glLibError("Cannot add; dimensions do not match!");
			#endif
			for(type_int i=0; i<m1->_rows; ++i)
				for(type_int j=0; j<m1->_columns; ++j)
					result->values[i][j] = m1->values[i][j] + m2->values[i][j];
		}
		static void subtract(type_matrix* m1, glLibMathMatrix<type_element,type_int>* m2, glLibMathMatrix<type_element,type_int>* result) {
			#ifdef GLLIB_ERROR_MODE_REPORT_ERRORS
			if (m1->_rows!=m2->_rows||m1->_columns!=m2->_columns) throw glLibError("Cannot add; dimensions do not match!");
			#endif
			for(type_int i=0; i<m1->_rows; ++i)
				for(type_int j=0; j<m1->_columns; ++j)
					result->values[i][j] = m1->values[i][j] - m2->values[i][j];
		}
		static void multiply(glLibMathMatrix<type_element,type_int>* m1, glLibMathMatrix<type_element,type_int>* m2, glLibMathMatrix<type_element,type_int>* result) {
			#ifdef GLLIB_ERROR_MODE_REPORT_ERRORS
			//http://mathworld.wolfram.com/MatrixMultiplication.html
			if (m1->_columns!=m2->_rows) {
				throw glLibError("Cannot multiply; dimensions are incompatible!");
			}
			if ((m1->_rows!=result->_rows)||(m2->_columns!=result->_columns)) {
				throw glLibError("Result matrix's dimensions are incorrect for the multiplication!");
			}
			#endif
			for (type_int i=0; i<m1->_rows; ++i) {
				for (type_int j=0; j<m2->_columns; ++j) {
					type_element sum = 0.0;
					for(type_int k=0; k<m1->_columns; ++k) { //m1->_columns should == m2->_rows
						sum += m1->values[i][k] * m2->values[k][j];
					}
					result->values[i][j] = sum;
				}
			}
		}
		static void transpose(glLibMathMatrix<type_element,type_int>* m) {
			#ifdef GLLIB_ERROR_MODE_REPORT_ERRORS
			if (m->_rows!=m->_columns) throw glLibError("Cannot transpose a matrix that is not square!");
			#endif

			for (type_int i=0; i<m->_rows; ++i) {
				for (type_int j=i; j<m->_columns; ++j) {
					type_element temp = m->values[i][j];
					m->values[i][j] = m->values[j][i];
					m->values[j][i] = temp;
				}
			}
		}
		static void transpose(glLibMathMatrix<type_element,type_int>* m, glLibMathMatrix<type_element,type_int>* result) {
			#ifdef GLLIB_ERROR_MODE_REPORT_ERRORS
			if (m==result) throw glLibError("This overload does not operate in-place.");
			if (m->_rows!=result->_columns||m->_columns!=result->_rows) throw glLibError("Matrices' dimensions incompatible!");
			if (m->_rows!=m->_columns) throw glLibError("Cannot transpose a matrix that is not square!");
			#endif
			for (type_int i=0; i<m->_rows; ++i) {
				for (type_int j=0; j<m->_columns; ++j) {
					result->values[i][j] = m->values[j][i];
				}
			}
		}
		//TODO: determinant function, minor function
		static inline type_element det(glLibMathMatrix<type_element,type_int>* matrix) { return determinant(matrix); }
		static type_element determinant(glLibMathMatrix<type_element,type_int>* matrix) {
			#ifdef GLLIB_ERROR_MODE_REPORT_ERRORS
			if (matrix->_rows!=matrix->_columns) throw glLibError("Cannot find the determinant of a non-square matrix!");
			#endif
			throw glLibError("Not implemented!");
		}

		/*static void minor(glLibMathMatrix<type_element,type_int>* matrix, type_int column, type_int row, glLibMathMatrix<type_element,type_int>* result) {
			throw glLibError("Not implemented!");
		}*/

		// in matrix A, add x times row r2 to row r1
		static void row_add_multiple(glLibMathMatrix<type_element,type_int>* m, type_element x, type_int r1, type_int r2) {
			for (type_int c=0; c<m->_columns; ++c) {
				m->values[r1][c] += x * m->values[r2][c];
			}
		}
		//TODO: swap column, column_add_multiple
		static void row_swap(glLibMathMatrix<type_element,type_int>* m, type_int r1, type_int r2) {
			type_element* temp = m->values[r1];
			m->values[r1] = m->values[r2];
			m->values[r2] = temp;
		}
		static void row_scale(glLibMathMatrix<type_element,type_int>* m, type_element sc, type_int row) {
			for (type_int c=0; c<m->_columns; ++c) {
				m->values[row][c] *= sc;
			}
		}
		//TODO: ref
		static void rref(glLibMathMatrix<type_element,type_int>* m) {
			type_int lead = 0;

			for (type_int r=0; r<m->_rows; ++r) {
				if (lead>=m->_columns) return;
				type_int i = r;
				while (m->values[i][lead]==0) {
					i++;
					if (i==m->_rows) {
						i = r;
						lead++;
						if (lead==m->_columns) {
							return;
						}
					}
				}
				row_swap(m, i,r);
				row_scale(m, 1.0/m->values[r][lead],r);
				for (i=0; i<m->_rows; ++i) {
					if (i!=r) {
						row_add_multiple(m, -m->values[i][lead], i, r);
					}
				}
				lead++;
			}
		}
		
		static void invert(glLibMathMatrix<type_element,type_int>* m, glLibMathMatrix<type_element,type_int>* result) {
			throw glLibError("Not implemented!");
		}
		
		static void copy(glLibMathMatrix<type_element,type_int>* src, glLibMathMatrix<type_element,type_int>* dst) {
			#ifdef GLLIB_ERROR_MODE_REPORT_ERRORS
			if ((src->_rows!=dst->_rows)||(src->_columns!=dst->_columns)) throw glLibError("Cannot copy; dimensions do not match!");
			#endif

			for(type_int i=0; i<src->_rows; ++i) {
				for(type_int j=0; j<src->_columns; ++j) {
					dst->values[i][j] = src->values[i][j];
				}
			}
		}
		static void copy(
			glLibMathMatrix<type_element,type_int>* src, glLibMathMatrix<type_element,type_int>* dst,
			type_int src_left, type_int src_right, type_int src_top, type_int src_bottom,
			type_int dst_left, type_int dst_right, type_int dst_top, type_int dst_bottom)
		{
			#ifdef GLLIB_ERROR_MODE_REPORT_ERRORS
			if (  src_left<0) throw glLibError(  "Source's left index cannot be negative!");
			if ( src_right<0) throw glLibError( "Source's right index cannot be negative!");
			if (   src_top<0) throw glLibError(   "Source's top index cannot be negative!");
			if (src_bottom<0) throw glLibError("Source's bottom index cannot be negative!");
			if (  dst_left<0) throw glLibError(  "Destination's left index cannot be negative!");
			if ( dst_right<0) throw glLibError( "Destination's right index cannot be negative!");
			if (   dst_top<0) throw glLibError(   "Destination's top index cannot be negative!");
			if (dst_bottom<0) throw glLibError("Destination's bottom index cannot be negative!");

			//TODO
			/*if (left> right) throw glLibError("Left index cannot be greater than the right index!");
			if ( top>bottom) throw glLibError("Top index cannot be greater than the bottom index!");

			if ( right+1>src->_columns) throw glLibError("Requested index is outside the bounds of the source matrix!");
			if (bottom+1>src->   _rows) throw glLibError("Requested index is outside the bounds of the source matrix!");

			if ( right-left+1>dst->_columns) throw glLibError("Requested submatrix will not fit in destination matrix!");
			if (bottom- top+1>dst->   _rows) throw glLibError("Requested submatrix will not fit in destination matrix!");

			if ((bottom-top+1!=dst->_rows)||(right-left+1!=dst->_columns)) throw glLibError("Cannot copy; dimensions do not match!");
			*/
			#endif

			for (type_int i1=src_top,i2=dst_top; i1<=src_bottom; ++i1,++i2) {
				for (type_int j1=src_left,j2=dst_left; j1<=src_right; ++j1,++j2) {
					//printf("Setting (%d,%d) to (%d,%d)\n",i2,j2,i1,j1);
					dst->values[i2][j2] = src->values[i1][j1];
				}
			}
		}
		
		static void scale(type_element scalar, glLibMathMatrix<type_element,type_int>* m) {
			for (type_int i=0; i<m->_rows; ++i)
				for (type_int j=0; j<m->_columns; ++j)
					m->values[i][j] = scalar * m->values[i][j];
		}
		static void scale(type_element scalar, glLibMathMatrix<type_element,type_int>* m, glLibMathMatrix<type_element,type_int>* result) {
			#ifdef GLLIB_ERROR_MODE_REPORT_ERRORS
			if (m->_rows!=result->_rows||m->_columns!=result->_columns) throw glLibError("Cannot scale; source and destination dimensions do not match!");
			#endif
			for (type_int i=0; i<m->_rows; ++i)
				for (type_int j=0; j<m->_columns; ++j)
					result->values[i][j] = scalar * m->values[i][j];
		}
		
		static void set_rotation_matrix_rad(glLibMathMatrix<type_element,type_int>* rot_matrix, type_element vec[3], type_element theta) {
			#ifdef GLLIB_ERROR_MODE_REPORT_ERRORS
			glLibInternal_check_3x3or4x4(rot_matrix);
			#endif
			type_element c = cos(theta);
			type_element s = sin(theta);
			type_element C = 1.0 - c;
			type_element x = vec[0]; type_element y = vec[1]; type_element z = vec[2];
			type_element xs = x*s;   type_element ys = y*s;   type_element zs = z*s;
			type_element xC = x*C;   type_element yC = y*C;   type_element zC = z*C;
			type_element xyC = x*yC; type_element yzC = y*zC; type_element zxC = z*xC;
			type_element** rot_matrix_values = rot_matrix->values;
			rot_matrix_values[0][0] = x*xC+c; rot_matrix_values[0][1] = xyC-zs; rot_matrix_values[0][2] = zxC+ys;
			rot_matrix_values[1][0] = xyC+zs; rot_matrix_values[1][1] = y*yC+c; rot_matrix_values[1][2] = yzC-xs;
			rot_matrix_values[2][0] = zxC-ys; rot_matrix_values[2][1] = yzC+xs; rot_matrix_values[2][2] = z*zC+c;
		}
		static inline void set_rotation_matrix_deg(glLibMathMatrix<type_element,type_int>* rot_matrix, type_element vec[3], type_element theta) {
			set_rotation_matrix_rad(rot_matrix,vec,radians(theta));
		}

		//Creates a matrix that transforms the unit i, j, and k vectors onto new vectors
		static void set_transform(glLibMathMatrix<type_element,type_int>* matrix, type_element new_i[3], type_element new_j[3], type_element new_k[3]) {
			//See http://www.gamedev.net/topic/604893-general-transformation-matrix/page__pid__4827470#entry4827470
			#ifdef GLLIB_ERROR_MODE_REPORT_ERRORS
			glLibInternal_check_4x4(matrix);
			#endif
			type_element** values = matrix->values;
			values[0][0] = new_i[0]; values[0][1] = new_j[0]; values[0][2] = new_k[0]; values[0][3] = 0;
			values[1][0] = new_i[1]; values[1][1] = new_j[1]; values[1][2] = new_k[1]; values[1][3] = 0;
			values[2][0] = new_i[2]; values[2][1] = new_j[2]; values[2][2] = new_k[2]; values[2][3] = 0;
			values[3][0] =        0; values[3][1] =        0; values[3][2] =        0; values[3][3] = 1;
		}

		static void conjugate_gradient(glLibMathMatrix<type_element,type_int>* A, glLibMathMatrix<type_element,type_int>* x, glLibMathMatrix<type_element,type_int>* b, int iterations) {
			#ifdef GLLIB_ERROR_MODE_REPORT_WARNINGS
			if (iterations>A->_rows) glLibWarning("Conjugate gradient should converge in no more than n iterations for A(n x n).");
			#endif

			glLibMathMatrix A_times_x(b->_rows,1);
			glLibMathMatrix A_times_p(b->_rows,1);

			glLibMathMatrix r(b->_rows,1);
			glLibMathMatrix r_next(b->_rows,1);

			glLibMathMatrix p(b->_rows,1);
			glLibMathMatrix alpha_p(b->_rows,1);
			glLibMathMatrix beta_p(b->_rows,1);

			multiply(A,x,&A_times_x);

			subtract(b,&A_times_x,&r);
			copy(&r,&p);

			for (int k=0;k<iterations;++k) {
				multiply(A,&p, &A_times_p);
				type_element alpha = dot(&r,&r) / dot(&p,&A_times_p);

				scale(alpha,&p, &alpha_p);
				add(x,&alpha_p, x);

				multiply(A,&p, &A_times_p);
				scale(alpha,&A_times_p);
				subtract(&r,&A_times_p, &r_next);

				//if r_next is sufficiently small then break

				type_element beta = dot(&r_next,&r_next) / dot(&r,&r);

				scale(beta,&p, &beta_p);
				add(&r_next,&beta_p, &p);

				copy(&r_next,&r);
			}
		}

		static type_element dot(glLibMathMatrix<type_element,type_int>* a, glLibMathMatrix<type_element,type_int>* b) {
			/*printf("Attempting to dot a:\n");
			a->print();
			printf("With b:\n");
			b->print();*/
			type_element result = 0.0;
			if (a->_columns==1) {
				if (b->_columns==1) {
					#ifdef GLLIB_ERROR_MODE_REPORT_ERRORS
					if (a->_rows!=b->_rows) goto ERROR_MATCH;
					#endif
					/*for (type_int i=0;i<a->_rows;++i) {
						type_element value1 = a->values[i][0];
						type_element value2 = b->values[i][0];
						type_element value = value1*value2;
						printf("Adding (a[%d,%d]=%f)*(b[%d,%d]=%f) = %f\n",0,i,value1,0,i,value2,value);
						result += value;
					}*/
					for (type_int i=0;i<a->_rows;++i) result += a->values[i][0]*b->values[i][0];
				}
				#ifdef GLLIB_ERROR_MODE_REPORT_ERRORS
				else if (b->_rows==1) {
				#else
				else {
				#endif
					#ifdef GLLIB_ERROR_MODE_REPORT_ERRORS
					if (a->_rows!=b->_columns) goto ERROR_MATCH;
					#endif
					for (type_int i=0;i<a->_rows;++i) result += a->values[i][0]*b->values[0][i];
				}
			}
			#ifdef GLLIB_ERROR_MODE_REPORT_ERRORS
			else if (a->_rows==1) {
			#else
			else {
			#endif
				if (b->_columns==1) {
					#ifdef GLLIB_ERROR_MODE_REPORT_ERRORS
					if (a->_columns!=b->_rows) goto ERROR_MATCH;
					#endif
					for (type_int i=0;i<a->_rows;++i) result += a->values[0][i]*b->values[i][0];
				}
				#ifdef GLLIB_ERROR_MODE_REPORT_ERRORS
				else if (b->_rows==1) {
				#else
				else {
				#endif
					#ifdef GLLIB_ERROR_MODE_REPORT_ERRORS
					if (a->_columns!=b->_columns) goto ERROR_MATCH;
					#endif
					for (type_int i=0;i<a->_rows;++i) result += a->values[0][i]*b->values[0][i];
				}
			}
			#ifdef GLLIB_ERROR_MODE_REPORT_ERRORS
			else goto ERROR_DIM;
			#endif

			return result;

			ERROR_MATCH: throw glLibError("Cannot compute dot product; dimensions incompatible!");
			ERROR_DIM: throw glLibError("Cannot computer dot product; matrix does not emulate a 2D array!");
		}
};

template <typename type_int> void glMultMatrixf(glLibMathMatrix<float,type_int>* matrix) {
	#ifdef GLLIB_ERROR_MODE_REPORT_ERRORS
	glLibInternal_check_4x4(matrix);
	#endif

	GLfloat values[16];
	char index = 0; //OpenGL is backwards from this, hence j, i.
	for (char i=0; i<4; ++i) { for (char j=0; j<4; ++j) { values[index++] = matrix->values[j][i]; } }

	glMultMatrixf(values);
}
template <typename type_int> void glMultMatrixd(glLibMathMatrix<double,type_int>* matrix) {
	#ifdef GLLIB_ERROR_MODE_REPORT_ERRORS
	glLibInternal_check_4x4(matrix);
	#endif

	GLdouble values[16];
	char index = 0; //OpenGL is backwards from this, hence j, i.
	for (char i=0; i<4; ++i) { for (char j=0; j<4; ++j) { values[index++] = matrix->values[j][i]; } }

	glMultMatrixd(values);
}